@@ -1,5 +1,11 @@
 #!/usr/bin/env python
 
+# Copyright Contributors to the OpenImageIO project.
+# SPDX-License-Identifier: Apache-2.0
+# https://github.com/AcademySoftwareFoundation/OpenImageIO
+
+from __future__ import print_function
+from __future__ import absolute_import
 import os
 import glob
 import sys
@@ -22,7 +28,7 @@ path = "../.."
 
 # Options for the command line
 parser = OptionParser()
-parser.add_option("-p", "--path", help="add to executable path",
+parser.add_option("-p", "--path", help="add to build area path",
                   action="store", type="string", dest="path", default="")
 parser.add_option("--devenv-config", help="use a MS Visual Studio configuration",
                   action="store", type="string", dest="devenv_config", default="")
@@ -37,29 +43,109 @@ if args and len(args) > 0 :
 if args and len(args) > 1 :
     path = args[1]
 path = os.path.normpath (path)
+OIIO_BUILD_ROOT = path
 
 tmpdir = "."
 tmpdir = os.path.abspath (tmpdir)
+redirect = " >> out.txt "
+wrapper_cmd = ""
+
+def make_relpath (path, start=os.curdir):
+    "Wrapper around os.path.relpath which always uses '/' as the separator."
+    p = os.path.relpath (path, start)
+    return p if platform.system() != 'Windows' else p.replace ('\\', '/')
+
+# Try to figure out where some key things are. Go by env variables set by
+# the cmake tests, but if those aren't set, assume somebody is running
+# this script by hand from inside build/testsuite/TEST and that
+# the rest of the tree has the standard layout.
+OIIO_TESTSUITE_ROOT = os.getenv('OIIO_TESTSUITE_ROOT', '')
+if OIIO_TESTSUITE_ROOT == '' :
+    if os.path.exists('../../../testsuite') :
+        OIIO_TESTSUITE_ROOT = '../../../testsuite'
+    elif os.path.exists('../../../../testsuite') :
+        OIIO_TESTSUITE_ROOT = '../../../../testsuite'
+OIIO_TESTSUITE_ROOT = make_relpath(OIIO_TESTSUITE_ROOT)
+OIIO_PROJECT_ROOT = make_relpath(OIIO_TESTSUITE_ROOT + "/..")
+
+OIIO_TESTSUITE_IMAGEDIR = os.getenv('OIIO_TESTSUITE_IMAGEDIR', '')
+if OIIO_TESTSUITE_IMAGEDIR == '' :
+    if os.path.exists('../oiio-images'):
+        OIIO_TESTSUITE_IMAGEDIR = '../oiio-images'
+OIIO_TESTSUITE_IMAGEDIR = make_relpath(OIIO_TESTSUITE_IMAGEDIR)
+# Set it back so tests can use it (python-imagebufalgo)
+os.putenv('OIIO_TESTSUITE_IMAGEDIR', OIIO_TESTSUITE_IMAGEDIR)
 
 refdir = "ref/"
 refdirlist = [ refdir ]
-parent = "../../../../../"
-test_source_dir = "../../../../testsuite/" + os.path.basename(os.path.abspath(srcdir))
+mytest = os.path.split(os.path.abspath(os.getcwd()))[-1]
+if str(mytest).endswith('.batch') :
+    mytest = mytest.split('.')[0]
+test_source_dir = os.getenv('OIIO_TESTSUITE_SRC',
+                            os.path.join(OIIO_TESTSUITE_ROOT, mytest))
+
+def oiio_app (app):
+    if (platform.system () != 'Windows' or options.devenv_config == ""):
+        cmd = os.path.join(OIIO_BUILD_ROOT, "bin", app) + " "
+    else:
+        cmd = os.path.join(OIIO_BUILD_ROOT, "bin", options.devenv_config, app) + " "
+    if wrapper_cmd != "":
+        cmd = wrapper_cmd + " " + cmd
+    return cmd
+
+
+# Ask oiiotool what version of OpenColorIO it has embedded
+ociover = subprocess.check_output([oiio_app('oiiotool').strip(),
+                                   '--echo', '{getattribute(opencolorio_version)}'])
+ociover = ociover.strip().decode('utf-8')[0:3]
+ociover = os.getenv('OCIO_VERSION_OVERRIDE', ociover)
+#print(f"OpenColorIO version = '{ociover}'")
+
+OCIO_env = os.getenv('OCIO')
+if OCIO_env is None:
+    os.environ['OCIO'] = 'ocio://default'
+    if ociover == '2.2' or ociover == '2.3' :
+        colorconfig_file = os.getenv('OIIO_TESTSUITE_OCIOCONFIG', 'ocio://default')
+    else :
+        colorconfig_file = os.getenv('OIIO_TESTSUITE_OCIOCONFIG',
+                                     '../common/OpenColorIO/nuke-default/config.ocio')
+    if not colorconfig_file.startswith('ocio://') :
+        colorconfig_file = make_relpath(colorconfig_file)
+
 
 command = ""
 outputs = [ "out.txt" ]    # default
+
+# The image comparison thresholds are tricky to remember. Here's the key:
+# A test fails if more than `failpercent` of pixel values differ by more
+# than `failthresh`, or if even one pixel differs by more than `hardfail`.
+failthresh = 0.004         # "Failure" threshold for any pixel value
+failpercent = 0.02         # Ok fo this percentage of pixels to "fail"
+hardfail = 0.012           # Even one pixel this wrong => hard failure
+allowfailures = 0          # Freebie failures
+
+# Some tests are designed for the app running to "fail" (in the sense of
+# terminating with an error return code), for example, a test that is designed
+# to present an error condition to check that it issues the right error. That
+# "failure" is a success of the test! For those cases, set `failureok = 1` to
+# indicate that the app having an error is fine, and the full test will pass
+# or fail based on comparing the output files.
 failureok = 0
-failthresh = 0.004
-hardfail = 0.012
-failpercent = 0.02
+
+
 anymatch = False
+cleanup_on_success = False
+if int(os.getenv('TESTSUITE_CLEANUP_ON_SUCCESS', '0')) :
+    cleanup_on_success = True
 
 image_extensions = [ ".tif", ".tx", ".exr", ".jpg", ".png", ".rla",
-                     ".dpx", ".iff", ".psd" ]
+                     ".dpx", ".iff", ".psd", ".bmp", ".fits", ".ico",
+                     ".jp2", ".sgi", ".tga", ".TGA", ".zfile" ]
 
 # print ("srcdir = " + srcdir)
 # print ("tmpdir = " + tmpdir)
-# print ("path = " + path)
+# print ("OIIO_BUILD_ROOT = " + OIIO_BUILD_ROOT)
+# print ("OIIO_TESTSUITE_IMAGEDIR = {} ({})".format(OIIO_TESTSUITE_IMAGEDIR, os.path.abspath(OIIO_TESTSUITE_IMAGEDIR)))
 # print ("refdir = " + refdir)
 # print ("test source dir = " + test_source_dir)
 
@@ -69,18 +155,31 @@ if platform.system() == 'Windows' :
     if os.path.exists (os.path.join (test_source_dir, "src")) and not os.path.exists("./src") :
         shutil.copytree (os.path.join (test_source_dir, "src"), "./src")
     # if not os.path.exists("../data") :
-    #     shutil.copytree ("../../../testsuite/data", "..")
+    #     shutil.copytree ("../../testsuite/data", "..")
     # if not os.path.exists("../common") :
-    #     shutil.copytree ("../../../testsuite/common", "..")
+    #     shutil.copytree ("../../testsuite/common", "..")
 else :
+    def newsymlink(src, dst):
+        print("newsymlink", src, dst)
+        # os.path.exists returns False for broken symlinks, so remove if thats the case
+        if os.path.islink(dst):
+            os.remove(dst)
+        os.symlink (src, dst)
     if not os.path.exists("./ref") :
-        os.symlink (os.path.join (test_source_dir, "ref"), "./ref")
+        newsymlink (os.path.join (test_source_dir, "ref"), "./ref")
     if os.path.exists (os.path.join (test_source_dir, "src")) and not os.path.exists("./src") :
-        os.symlink (os.path.join (test_source_dir, "src"), "./src")
+        newsymlink (os.path.join (test_source_dir, "src"), "./src")
     if not os.path.exists("./data") :
-        os.symlink (test_source_dir, "./data")
-    if not os.path.exists("../common") :
-        os.symlink ("../../../testsuite/common", "../common")
+        newsymlink (test_source_dir, "./data")
+
+
+if os.getenv("Python_EXECUTABLE") :
+    pythonbin = os.getenv("Python_EXECUTABLE")
+else :
+    pythonbin = 'python'
+    if os.getenv("PYTHON_VERSION") :
+        pythonbin += os.getenv("PYTHON_VERSION")
+#print ("pythonbin = ", pythonbin)
 
 
 ###########################################################################
@@ -96,8 +195,10 @@ def text_diff (fromfile, tofile, diff_file=None):
     try:
         fromdate = time.ctime (os.stat (fromfile).st_mtime)
         todate = time.ctime (os.stat (tofile).st_mtime)
-        fromlines = open (fromfile, 'rU').readlines()
-        tolines   = open (tofile, 'rU').readlines()
+        fromlines = open (fromfile, 'r').readlines()
+        tolines   = open (tofile, 'r').readlines()
+        # if replace_relative:
+        #     tolines = replace_relative(tolines)
     except:
         print ("Unexpected error:", sys.exc_info()[0])
         return -1
@@ -121,36 +222,39 @@ def text_diff (fromfile, tofile, diff_file=None):
     return 1
 
 
-
-def oiio_relpath (path, start=os.curdir):
-    "Wrapper around os.path.relpath which always uses '/' as the separator."
-    p = os.path.relpath (path, start)
-    return p if sys.platform != "win32" else p.replace ('\\', '/')
-
-
-def oiio_app (app):
-    # When we use Visual Studio, built applications are stored
-    # in the app/$(OutDir)/ directory, e.g., Release or Debug.
-    if (platform.system () != 'Windows' or options.devenv_config == ""):
-        return os.path.join (path, "src", app, app) + " "
-    else:
-        return os.path.join (path, "src", app, options.devenv_config, app) + " "
+def run_app(app, silent=False, concat=True):
+    command = app
+    if not silent:
+        command += redirect
+    if concat:
+        command += " ;\n"
+    return command
 
 
 # Construct a command that will print info for an image, appending output to
 # the file "out.txt".  If 'safematch' is nonzero, it will exclude printing
 # of fields that tend to change from run to run or release to release.
 def info_command (file, extraargs="", safematch=False, hash=True,
-                  verbose=True) :
-    args = "--info"
+                  verbose=True, silent=False, concat=True, failureok=False,
+                  info_program="oiiotool") :
+    args = ""
+    if info_program == "oiiotool" :
+        args += " --info"
     if verbose :
         args += " -v -a"
     if safematch :
         args += " --no-metamatch \"DateTime|Software|OriginatingProgram|ImageHistory\""
     if hash :
         args += " --hash"
-    return (oiio_app("oiiotool") + args + " " + extraargs
-            + " " + oiio_relpath(file,tmpdir) + " >> out.txt ;\n")
+    cmd = (oiio_app(info_program) + args + " " + extraargs
+            + " " + make_relpath(file,tmpdir))
+    if not silent :
+        cmd += redirect
+    if failureok :
+        cmd += " || true "
+    if concat:
+        cmd += " ;\n"
+    return cmd
 
 
 # Construct a command that will compare two images, appending output to
@@ -162,12 +266,13 @@ def diff_command (fileA, fileB, extraargs="", silent=False, concat=True) :
                + " -fail " + str(failthresh)
                + " -failpercent " + str(failpercent)
                + " -hardfail " + str(hardfail)
+               + " -allowfailures " + str(allowfailures)
                + " -warn " + str(2*failthresh)
                + " -warnpercent " + str(failpercent)
-               + " " + extraargs + " " + oiio_relpath(fileA,tmpdir) 
-               + " " + oiio_relpath(fileB,tmpdir))
+               + " " + extraargs + " " + make_relpath(fileA,tmpdir)
+               + " " + make_relpath(fileB,tmpdir))
     if not silent :
-        command += " >> out.txt"
+        command += redirect
     if concat:
         command += " ;\n"
     return command
@@ -178,12 +283,12 @@ def diff_command (fileA, fileB, extraargs="", silent=False, concat=True) :
 def maketx_command (infile, outfile, extraargs="",
                     showinfo=False, showinfo_extra="",
                     silent=False, concat=True) :
-    command = (oiio_app("maketx") 
-               + " " + oiio_relpath(infile,tmpdir) 
+    command = (oiio_app("maketx")
+               + " " + make_relpath(infile,tmpdir)
                + " " + extraargs
-               + " -o " + oiio_relpath(outfile,tmpdir) )
+               + " -o " + make_relpath(outfile,tmpdir))
     if not silent :
-        command += " >> out.txt"
+        command += redirect
     if concat:
         command += " ;\n"
     if showinfo:
@@ -201,7 +306,7 @@ def maketx_command (infile, outfile, extraargs="",
 def rw_command (dir, filename, testwrite=True, use_oiiotool=False, extraargs="",
                 preargs="", idiffextraargs="", output_filename="",
                 safematch=False, printinfo=True) :
-    fn = oiio_relpath (dir + "/" + filename, tmpdir)
+    fn = make_relpath (dir + "/" + filename, tmpdir)
     if printinfo :
         cmd = info_command (fn, safematch=safematch)
     else :
@@ -211,31 +316,51 @@ def rw_command (dir, filename, testwrite=True, use_oiiotool=False, extraargs="",
     if testwrite :
         if use_oiiotool :
             cmd = (cmd + oiio_app("oiiotool") + preargs + " " + fn
-                   + " " + extraargs + " -o " + output_filename + " >> out.txt ;\n")
+                   + " " + extraargs + " -o " + output_filename + redirect + ";\n")
         else :
             cmd = (cmd + oiio_app("iconvert") + preargs + " " + fn
-                   + " " + extraargs + " " + output_filename + " >> out.txt ;\n")
+                   + " " + extraargs + " " + output_filename + redirect + ";\n")
         cmd = (cmd + oiio_app("idiff") + " -a " + fn
                + " -fail " + str(failthresh)
                + " -failpercent " + str(failpercent)
                + " -hardfail " + str(hardfail)
+               + " -allowfailures " + str(allowfailures)
                + " -warn " + str(2*failthresh)
-               + " " + idiffextraargs + " " + output_filename + " >> out.txt ;\n")
+               + " " + idiffextraargs + " " + output_filename + redirect + ";\n")
     return cmd
 
 
 # Construct a command that will testtex
-def testtex_command (file, extraargs="") :
-    cmd = (oiio_app("testtex") + " " + file + " " + extraargs + " " +
-           " >> out.txt ;\n")
+def testtex_command (file, extraargs="", silent=False, concat=True) :
+    cmd = oiio_app("testtex") + " " + file + " " + extraargs + " "
+    if not silent :
+        cmd += redirect
+    if concat:
+        cmd += " ;\n"
+    return cmd
+
+
+# Construct a command that will run iconvert and append its output to out.txt
+def iconvert (args, silent=False, concat=True, failureok=False) :
+    cmd = (oiio_app("iconvert") + " " + args)
+    if not silent :
+        cmd += redirect
+    if failureok :
+        cmd += " || true "
+    if concat:
+        cmd += " ;\n"
     return cmd
 
 
 # Construct a command that will run oiiotool and append its output to out.txt
-def oiiotool (args, silent=False, concat=True) :
-    cmd = (oiio_app("oiiotool") + " " + args)
+def oiiotool (args, silent=False, concat=True, failureok=False) :
+    cmd = (oiio_app("oiiotool") + " "
+           + args)
+    #       + "-colorconfig " + colorconfig_file + " "
     if not silent :
-        cmd += " >> out.txt"
+        cmd += redirect
+    if failureok :
+        cmd += " || true "
     if concat:
         cmd += " ;\n"
     return cmd
@@ -261,10 +386,11 @@ def checkref (name, refdirlist) :
             pattern = "*.*"
         else :
             pattern = prefix+"-*"+extension+"*"
+        print("comparisons are", ([defaulttest] + glob.glob (os.path.join (ref, pattern))))
         for testfile in ([defaulttest] + glob.glob (os.path.join (ref, pattern))) :
             if not os.path.exists(testfile) :
                 continue
-            # print ("comparing " + name + " to " + testfile)
+            print ("comparing " + name + " to " + testfile)
             if extension in image_extensions :
                 # images -- use idiff
                 cmpcommand = diff_command (name, testfile, concat=False, silent=True)
@@ -293,6 +419,9 @@ def runtest (command, outputs, failureok=0) :
 #    print ("working dir = " + tmpdir)
     os.chdir (srcdir)
     open ("out.txt", "w").close()    # truncate out.txt
+    open ("out.err.txt", "w").close()    # truncate out.txt
+    if os.path.isfile("debug.log") :
+        os.remove ("debug.log")
 
     if options.path != "" :
         sys.path = [options.path] + sys.path
@@ -317,12 +446,21 @@ def runtest (command, outputs, failureok=0) :
 
     for out in outputs :
         (prefix, extension) = os.path.splitext(out)
+        # On Windows, change line endings of text files to unix style before
+        # comparison to reference output.
+        if (platform.system() == 'Windows' and os.path.exists(out)
+                and extension == '.txt') :
+            os.rename (out, "crlf.txt")
+            os.system ("tr -d '\\r' < crlf.txt > " + out)
+            if os.path.exists('crlf.txt') :
+                os.remove('crlf.txt')
+
         (ok, testfile) = checkref (out, refdirlist)
 
         if ok :
             if extension in image_extensions :
                 # If we got a match for an image, save the idiff results
-                os.system (diff_command (out, testfile, silent=False))
+                os.system (diff_command (out, testfile, silent=False, concat=False))
             print ("PASS: " + out + " matches " + testfile)
         else :
             err = 1
@@ -333,13 +471,22 @@ def runtest (command, outputs, failureok=0) :
                 # file and the diff, for easy debugging.
                 print ("-----" + out + "----->")
                 print (open(out,'r').read() + "<----------")
+                print ("-----" + testfile + "----->")
+                print (open(testfile,'r').read() + "<----------")
+                os.system ("ls -al " +out+" "+testfile)
                 print ("Diff was:\n-------")
-                print (open (out+".diff", 'rU').read())
+                print (open (out+".diff", 'r').read())
             if extension in image_extensions :
                 # If we failed to get a match for an image, send the idiff
                 # results to the console
-                os.system (diff_command (out, testfile, silent=False))
-
+                os.system (diff_command (out, testfile, silent=False, concat=False))
+            if os.path.isfile("debug.log") and os.path.getsize("debug.log") :
+                print ("---   DEBUG LOG   ---\n")
+                #flog = open("debug.log", "r")
+                # print (flog.read())
+                with open("debug.log", "r") as flog :
+                    print (flog.read())
+                print ("--- END DEBUG LOG ---\n")
     return (err)
 
 
@@ -356,10 +503,8 @@ with open(os.path.join(test_source_dir,"run.py")) as f:
     exec (code)
 
 # Allow a little more slop for slight pixel differences when in DEBUG
-# mode or when running on remote Travis-CI or Appveyor machines.
-if (("TRAVIS" in os.environ and os.environ["TRAVIS"]) or
-    ("APPVEYOR" in os.environ and os.environ["APPVEYOR"]) or
-    ("DEBUG" in os.environ and os.environ["DEBUG"])) :
+# mode or when running on remote CI machines.
+if (os.getenv('CI') or os.getenv('DEBUG')) :
     failthresh *= 2.0
     hardfail *= 2.0
     failpercent *= 2.0
@@ -367,4 +512,11 @@ if (("TRAVIS" in os.environ and os.environ["TRAVIS"]) or
 
 # Run the test and check the outputs
 ret = runtest (command, outputs, failureok=failureok)
+
+if ret == 0 and cleanup_on_success :
+    for ext in image_extensions + [ ".txt", ".diff" ] :
+        for f in glob.iglob (srcdir + '/*' + ext) :
+            os.remove(f)
+            #print('REMOVED ', f)
+
 sys.exit (ret)
